// LzBinTree.cs
//  
//  Copyright (C) 2009,2011 Bartek Jasicki
// 
//  This file is part of Grubng
// 
//  Grubng is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

using System;

namespace Grubng
{
	internal sealed class BinTree : InWindow, IMatchFinder
	{
		uint _cyclicBufferPos;
		uint _cyclicBufferSize = 0;
		uint _matchMaxLen;

		uint[] _son;
		uint[] _hash;

		uint _cutValue = 255;
		uint _hashMask;
		uint _hashSizeSum = 0;

		bool HASH_ARRAY = true;
	
		uint kNumHashDirectBytes = 0;
		uint kMinMatchCheck = 4;
		uint kFixHashSize = 66560;
		static readonly uint[] Table = new uint[256];
		
		static void CRC()
		{
			uint r;
			for (uint i = 0; i < 256; i++)
			{
				r = i;
				for (int j = 0; j < 8; j++)
					if ((r & 1) != 0)
						r = (r >> 1) ^ 3988292384;
					else
						r >>= 1;
				Table[i] = r;
			}
		}
		
		public void SetType(int numHashBytes)
		{
			HASH_ARRAY = (numHashBytes > 2);
			if (HASH_ARRAY)
			{
				kNumHashDirectBytes = 0;
				kMinMatchCheck = 4;
				kFixHashSize = 66560;
			}
			else
			{
				kNumHashDirectBytes = 2;
				kMinMatchCheck = 2 + 1;
				kFixHashSize = 0;
			}
		}

		public new void SetStream(System.IO.Stream stream) { base.SetStream(stream); }
		public new void ReleaseStream() { base.ReleaseStream(); }
		
		public new void Init()
		{
			base.Init();
			for (uint i = 0; i < _hashSizeSum; i++)
				_hash[i] = 0;
			_cyclicBufferPos = 0;
			ReduceOffsets(-1);
			BinTree.CRC();
		}

		public new void MovePos()
		{
			if (++_cyclicBufferPos >= _cyclicBufferSize)
				_cyclicBufferPos = 0;
			base.MovePos();
			if (pos == 2147483647)
				Normalize();
		}

		public new Byte GetIndexByte(int index) { return base.GetIndexByte(index); }

		public new uint GetMatchLen(int index, uint distance, uint limit)
		{ return base.GetMatchLen(index, distance, limit); }

		public new uint GetNumAvailableBytes() 
		{ 
			return base.GetNumAvailableBytes;
		}

		public void Create(uint historySize, uint keepAddBufferBefore,
				uint matchMaxLen, uint keepAddBufferAfter)
		{
			if (historySize > 2147481087)
				throw new ArgumentException("Invalid argument");
			_cutValue = 16 + (matchMaxLen >> 1);
				
			uint windowReservSize = (historySize + keepAddBufferBefore +
					matchMaxLen + keepAddBufferAfter) / 2 + 256;

			base.Create(historySize + keepAddBufferBefore, matchMaxLen + keepAddBufferAfter, windowReservSize);

			_matchMaxLen = matchMaxLen;

			uint cyclicBufferSize = historySize + 1;
			if (_cyclicBufferSize != cyclicBufferSize)
				_son = new uint[(_cyclicBufferSize = cyclicBufferSize) * 2];

			uint hs = 65536;

			if (HASH_ARRAY)
			{
				hs = historySize - 1;
				hs |= (hs >> 1);
				hs |= (hs >> 2);
				hs |= (hs >> 4);
				hs |= (hs >> 8);
				hs >>= 1;
				hs |= 65535;
				if (hs > 16777216)
					hs >>= 1;
				_hashMask = hs;
				hs++;
				hs += kFixHashSize;
			}
			if (hs != _hashSizeSum)
				_hash = new uint[_hashSizeSum = hs];
		}

		public uint GetMatches(uint[] distances)
		{
			uint lenLimit;
			if (pos + _matchMaxLen <= streamPos)
				lenLimit = _matchMaxLen;
			else
			{
				lenLimit = streamPos - pos;
				if (lenLimit < kMinMatchCheck)
				{
					MovePos();
					return 0;
				}
			}

			uint offset = 0;
			uint matchMinPos = (pos > _cyclicBufferSize) ? (pos - _cyclicBufferSize) : 0;
			uint cur = bufferOffset + pos;
			uint maxLen = 1; // to avoid items for len < hashSize;
			uint hashValue, hash2Value = 0, hash3Value = 0;

			if (HASH_ARRAY)
			{
				uint temp = Table[bufferBase[cur]] ^ bufferBase[cur + 1];
				hash2Value = temp & (1023);
				temp ^= ((uint)(bufferBase[cur + 2]) << 8);
				hash3Value = temp & (65535);
				hashValue = (temp ^ (Table[bufferBase[cur + 3]] << 5)) & _hashMask;
			}
			else
				hashValue = bufferBase[cur] ^ ((uint)(bufferBase[cur + 1]) << 8);

			uint curMatch = _hash[kFixHashSize + hashValue];
			if (HASH_ARRAY)
			{
				uint curMatch2 = _hash[hash2Value];
				uint curMatch3 = _hash[1024 + hash3Value];
				_hash[hash2Value] = pos;
				_hash[1024 + hash3Value] = pos;
				if (curMatch2 > matchMinPos)
					if (bufferBase[bufferOffset + curMatch2] == bufferBase[cur])
					{
						distances[offset++] = maxLen = 2;
						distances[offset++] = pos - curMatch2 - 1;
					}
				if (curMatch3 > matchMinPos)
					if (bufferBase[bufferOffset + curMatch3] == bufferBase[cur])
					{
						if (curMatch3 == curMatch2)
							offset -= 2;
						distances[offset++] = maxLen = 3;
						distances[offset++] = pos - curMatch3 - 1;
						curMatch2 = curMatch3;
					}
				if (offset != 0 && curMatch2 == curMatch)
				{
					offset -= 2;
					maxLen = 1;
				}
			}

			_hash[kFixHashSize + hashValue] = pos;

			uint ptr0 = (_cyclicBufferPos << 1) + 1;
			uint ptr1 = (_cyclicBufferPos << 1);

			uint len0, len1;
			len0 = len1 = kNumHashDirectBytes;
			
			if (kNumHashDirectBytes != 0)
			{
				if (curMatch > matchMinPos)
				{
					if (bufferBase[bufferOffset + curMatch + kNumHashDirectBytes] !=
							bufferBase[cur + kNumHashDirectBytes])
					{
						distances[offset++] = maxLen = kNumHashDirectBytes;
						distances[offset++] = pos - curMatch - 1;
					}
				}
			}
			
			uint count = _cutValue;
			
			uint delta;
			uint cyclicPos;
			uint pby1;
			uint len;
			while(true)
			{
				if(curMatch <= matchMinPos || count-- == 0)
				{
					_son[ptr0] = _son[ptr1] = 0;
					break;
				}
				delta = pos - curMatch;
				cyclicPos = ((delta <= _cyclicBufferPos) ?
							(_cyclicBufferPos - delta) :
							(_cyclicBufferPos - delta + _cyclicBufferSize)) << 1;

				pby1 = bufferOffset + curMatch;
				len = Math.Min(len0, len1);
				if (bufferBase[pby1 + len] == bufferBase[cur + len])
				{
					while(++len != lenLimit)
						if (bufferBase[pby1 + len] != bufferBase[cur + len])
							break;
					if (maxLen < len)
					{
						distances[offset++] = maxLen = len;
						distances[offset++] = delta - 1;
						if (len == lenLimit)
						{
							_son[ptr1] = _son[cyclicPos];
							_son[ptr0] = _son[cyclicPos + 1];
							break;
						}
					}
				}
				if (bufferBase[pby1 + len] < bufferBase[cur + len])
				{
					_son[ptr1] = curMatch;
					ptr1 = cyclicPos + 1;
					curMatch = _son[ptr1];
					len1 = len;
				}
				else
				{
					_son[ptr0] = curMatch;
					ptr0 = cyclicPos;
					curMatch = _son[ptr0];
					len0 = len;
				}
			}
			MovePos();
			return offset;
		}

		public void Skip(uint num)
		{
			uint lenLimit;
			uint matchMinPos;
			uint cur;
			uint hashValue;
			uint temp;
			uint hash2Value;
			uint hash3Value;
			uint curMatch;
			uint ptr0;
			uint ptr1;
			uint len0, len1;
			uint count;
			uint delta;
			uint cyclicPos;
			uint pby1;
			uint len;
			do
			{
				lenLimit = 0;
				if (pos + _matchMaxLen <= streamPos)
					lenLimit = _matchMaxLen;
				else
				{
					lenLimit = streamPos - pos;
					if (lenLimit < kMinMatchCheck)
					{
						MovePos();
						continue;
					}
				}

				matchMinPos = (pos > _cyclicBufferSize) ? (pos - _cyclicBufferSize) : 0;
				cur = bufferOffset + pos;

				hashValue = 0;

				if (HASH_ARRAY)
				{
					temp = Table[bufferBase[cur]] ^ bufferBase[cur + 1];
					hash2Value = temp & (1023);
					_hash[hash2Value] = pos;
					temp ^= ((uint)(bufferBase[cur + 2]) << 8);
					hash3Value = temp & (65535);
					_hash[1024 + hash3Value] = pos;
					hashValue = (temp ^ (Table[bufferBase[cur + 3]] << 5)) & _hashMask;
				}
				else
					hashValue = bufferBase[cur] ^ ((uint)(bufferBase[cur + 1]) << 8);

				curMatch = _hash[kFixHashSize + hashValue];
				_hash[kFixHashSize + hashValue] = pos;

				ptr0 = (_cyclicBufferPos << 1) + 1;
				ptr1 = (_cyclicBufferPos << 1);

				len0 = len1 = kNumHashDirectBytes;

				count = _cutValue;
				while (true)
				{
					if (curMatch <= matchMinPos || count-- == 0)
					{
						_son[ptr0] = _son[ptr1] = 0;
						break;
					}

					delta = pos - curMatch;
					cyclicPos = ((delta <= _cyclicBufferPos) ?
								(_cyclicBufferPos - delta) :
								(_cyclicBufferPos - delta + _cyclicBufferSize)) << 1;

					pby1 = bufferOffset + curMatch;
					len = Math.Min(len0, len1);
					if (bufferBase[pby1 + len] == bufferBase[cur + len])
					{
						while (++len != lenLimit)
							if (bufferBase[pby1 + len] != bufferBase[cur + len])
								break;
						if (len == lenLimit)
						{
							_son[ptr1] = _son[cyclicPos];
							_son[ptr0] = _son[cyclicPos + 1];
							break;
						}
					}
					if (bufferBase[pby1 + len] < bufferBase[cur + len])
					{
						_son[ptr1] = curMatch;
						ptr1 = cyclicPos + 1;
						curMatch = _son[ptr1];
						len1 = len;
					}
					else
					{
						_son[ptr0] = curMatch;
						ptr0 = cyclicPos;
						curMatch = _son[ptr0];
						len0 = len;
					}
				}
				MovePos();
			}
			while (--num != 0);
		}

		static void NormalizeLinks(uint[] items, uint numItems, uint subValue)
		{
			uint value;
			for (uint i = 0; i < numItems; i++)
			{
				value = items[i];
				if (value <= subValue)
					value = 0;
				else
					value -= subValue;
				items[i] = value;
			}
		}

		void Normalize()
		{
			uint subValue = pos - _cyclicBufferSize;
			NormalizeLinks(_son, _cyclicBufferSize * 2, subValue);
			NormalizeLinks(_hash, _hashSizeSum, subValue);
			ReduceOffsets((int)subValue);
		}

		/// <summary>
		/// Function override InWindow.Dispose() function
		/// </summary>
		/// <param name="disposing">
		/// A <see cref="System.Boolean"/> If true, dispose base class
		/// </param>
		protected override void Dispose(bool disposing)
		{
			if (!Disposed)
			{
				this._son = null;
				this._hash = null;
				base.Dispose(disposing);
			}
		}    
	}
}
