// LzInWindow.cs
//  
//  Copyright (C) 2009 Bartek Jasicki based on LZMA SDK
// 
//  This file is part of Grubng
// 
//  Grubng is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

using System;

namespace Grubng
{
	internal class InWindow : IDisposable
	{
		public Byte[] bufferBase = null; // pointer to buffer with data
		System.IO.Stream _stream;
		uint _posLimit; // offset (from _buffer) of first byte when new block reading must be done
		bool _streamEndWasReached; // if (true) then streamPos shows real end of stream

		uint _pointerToLastSafePosition;
		
		/// <summary>
		/// Check if resources are disposed.
		/// </summary>
		protected bool Disposed {get; private set;}

		public uint bufferOffset;

		public uint blockSize; // Size of Allocated memory block
		public uint pos; // offset (from _buffer) of curent byte
		uint _keepSizeBefore; // how many BYTEs must be kept in buffer before pos
		uint _keepSizeAfter; // how many BYTEs must be kept buffer after pos
		public uint streamPos; // offset (from _buffer) of first not read byte from Stream

		void MoveBlock()
		{
			this.DisposeCheck();
			uint offset = (uint)(bufferOffset) + pos - _keepSizeBefore;
			// we need one additional byte, since MovePos moves on 1 byte.
			if (offset > 0)
				offset--;
			
			uint numBytes = (uint)(bufferOffset) + streamPos - offset;

			// check negative offset ????
			for (uint i = 0; i < numBytes; i++)
				bufferBase[i] = bufferBase[offset + i];
			bufferOffset -= offset;
		}

		protected virtual void ReadBlock()
		{
			this.DisposeCheck();
			if (_streamEndWasReached)
				return;
			int size;
			int numReadBytes;
			uint pointerToPostion;
			while (true)
			{
				size = (int)((0 - bufferOffset) + blockSize - streamPos);
				if (size == 0)
					return;
				numReadBytes = _stream.Read(bufferBase, (int)(bufferOffset + streamPos), size);
				if (numReadBytes == 0)
				{
					_posLimit = streamPos;
					pointerToPostion = bufferOffset + _posLimit;
					if (pointerToPostion > _pointerToLastSafePosition)
						_posLimit = (uint)(_pointerToLastSafePosition - bufferOffset);

					_streamEndWasReached = true;
					return;
				}
				streamPos += (uint)numReadBytes;
				if (streamPos >= pos + _keepSizeAfter)
					_posLimit = streamPos - _keepSizeAfter;
			}
		}

		void Free() 
		{
			this.DisposeCheck();
			bufferBase = null; 
		}

		protected void Create(uint keepSizeBefore, uint keepSizeAfter, uint keepSizeReserv)
		{
			this.DisposeCheck();
			_keepSizeBefore = keepSizeBefore;
			_keepSizeAfter = keepSizeAfter;
			uint blockSize2 = keepSizeBefore + keepSizeAfter + keepSizeReserv;
			if (bufferBase == null || blockSize != blockSize2)
			{
				Free();
				blockSize = blockSize2;
				bufferBase = new Byte[blockSize];
			}
			_pointerToLastSafePosition = blockSize - keepSizeAfter;
		}

		protected void SetStream(System.IO.Stream stream) 
		{
			this.DisposeCheck();
			_stream = stream; 
		}
		
		protected void ReleaseStream() 
		{
			this.DisposeCheck();
			_stream.Close(); 
		}

		protected void Init()
		{
			this.DisposeCheck();
			bufferOffset = 0;
			pos = 0;
			streamPos = 0;
			_streamEndWasReached = false;
			ReadBlock();
		}

		protected void MovePos()
		{
			this.DisposeCheck();
			pos++;
			if (pos > _posLimit)
			{
				uint pointerToPostion = bufferOffset + pos;
				if (pointerToPostion > _pointerToLastSafePosition)
					MoveBlock();
				ReadBlock();
			}
		}

		protected byte GetIndexByte(Int32 index) 
		{
			this.DisposeCheck();
			return bufferBase[bufferOffset + pos + index]; 
		}

		// index + limit have not to exceed _keepSizeAfter;
		protected uint GetMatchLen(Int32 index, uint distance, uint limit)
		{
			this.DisposeCheck();
			if (_streamEndWasReached)
				if ((pos + index) + limit > streamPos)
					limit = streamPos - (uint)(pos + index);
			distance++;
			uint pby = bufferOffset + pos + (uint)index;

			uint i;
			for (i = 0; i < limit && bufferBase[pby + i] == bufferBase[pby + i - distance]; i++);
			return i;
		}

		protected uint GetNumAvailableBytes
		{
			get {return streamPos - pos; }
		}

		protected void ReduceOffsets(Int32 subValue)
		{
			this.DisposeCheck();
			bufferOffset += (uint)subValue;
			_posLimit -= (uint)subValue;
			pos -= (uint)subValue;
			streamPos -= (uint)subValue;
		}
		
		/// <summary>
		/// Check for dispose. If was called, throw exception
		/// </summary>
		void DisposeCheck()
		{
			if (this.Disposed)
			{
				throw new ObjectDisposedException(GetType().Name);
			}
		}
		
		/// <summary>
		/// Function dispose unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			this.Dispose(true);
			GC.SuppressFinalize(this);
		}
		
		/// <summary>
		/// Function dispose unmanaged resources.
		/// </summary>
		/// <param name="disposing">
		/// A <see cref="System.Boolean"/> If true, dispose unmanaged resources.
		/// </param>
		protected virtual void Dispose(bool disposing)
		{
			if (!Disposed)
			{
				if (disposing)
				{
					if (this._stream != null)
					{
						this._stream.Dispose();
					}
					this.bufferBase = null;
				}
				Disposed = true;
			}
		}
	}
}
