// LzmaBase.cs
//  
//  Copyright (C) 2009 Bartek Jasicki based on LZMA SDK
// 
//  This file is part of Grubng
// 
//  Grubng is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

using System;

namespace Grubng
{
	internal static class Base
	{
		public const uint kNumRepDistances = 4;
		public const uint kNumStates = 12;

		public struct State : IEquatable<State>
		{
			public uint Index;
			public void Init() { Index = 0; }
			public void UpdateChar()
			{
				if (Index < 4) Index = 0;
				else if (Index < 10) Index -= 3;
				else Index -= 6;
			}
				
			/// <summary>
			/// Merge UpdateMatch(), UpdateRep(), UpdateShortRep()
			/// </summary>
			/// <param name="type">
			/// A <see cref="System.Int16"/>
			/// 1 - Match, 2 - Rep, 3 - ShortRep
			/// </param>
			public void Update(short type)
			{
				switch (type)
				{
				case 1:
					Index = (uint)(Index < 7 ? 7 : 10);
					break;
				case 2:
					Index = (uint)(Index < 7 ? 8 : 11);
					break;
				case 3:
					Index = (uint)(Index < 7 ? 9 : 11);
					break;
				default:
					break;
				}
			}
			
			public bool IsCharState 
			{ 
				get {return Index < 7; }
			}
			
			/// <summary>
			/// Function override standard Object.Equals(object) function
			/// </summary>
			/// <param name="obj">
			/// A <see cref="System.Object"/> Object to compare
			/// </param>
			/// <returns>
			/// A <see cref="System.Boolean"/> True if both objects are that same, otherwise false
			/// </returns>
			public override bool Equals(object obj)
			{
				if (obj == null)
				{
					return false;
				}
				if (GetType() != obj.GetType()) 
				{
					return false;
				}    
				State state = (State)obj;                    
				return this == state;
			}
        
			/// <summary>
			/// Function overload function State.Equals(object)
			/// </summary>
			/// <param name="right">
			/// A <see cref="State"/> value to compare
			/// </param>
			/// <returns>
			/// A <see cref="System.Boolean"/> True if both objects are that same, otherwise false
			/// </returns>
			public bool Equals(State right)
			{                    
				return this == right;
			}
 
			/// <summary>
			/// Function override operator ==
			/// </summary>
			/// <param name="left">
			/// A <see cref="State"/> left value to compare
			/// </param>
			/// <param name="right">
			/// A <see cref="State"/> right value to compare
			/// </param>
			/// <returns>
			/// A <see cref="System.Boolean"/> True if both object are that same, otherwise false
			/// </returns>
			public static bool operator == (State left, State right)
			{
				return left.Equals(right);
			}
    
			/// <summary>
			/// Function override operator !=
			/// </summary>
			/// <param name="left">
			/// A <see cref="State"/> left value to compare
			/// </param>
			/// <param name="right">
			/// A <see cref="State"/> right value to compare
			/// </param>
			/// <returns>
			/// A <see cref="System.Boolean"/> True if both object different, otherwise false
			/// </returns>
			public static bool operator != (State left, State right)
			{
				return !left.Equals(right);
			}
    
			/// <summary>
			/// Function override function Object.GetHashCode();
			/// </summary>
			/// <returns>
			/// A <see cref="System.Int32"/> hash code for current instance
			/// </returns>
			public override int GetHashCode()
			{
				int hash; 
				unchecked
				{
					hash = Index.GetHashCode();
				}       
				return hash;
			}
		}

		public const int kNumPosSlotBits = 6;
		public const int kDicLogSizeMin = 0;

		public const int kNumLenToPosStatesBits = 2; // it's for speed optimization
		public const uint kNumLenToPosStates = 1 << kNumLenToPosStatesBits;

		public const uint kMatchMinLen = 2;

		public static uint GetLenToPosState(uint len)
		{
			len -= kMatchMinLen;
			if (len < kNumLenToPosStates)
				return len;
			return (uint)(kNumLenToPosStates - 1);
		}

		public const int kNumAlignBits = 4;
		public const uint kAlignTableSize = 1 << kNumAlignBits;
		public const uint kAlignMask = (kAlignTableSize - 1);

		public const uint kStartPosModelIndex = 4;
		public const uint kEndPosModelIndex = 14;
		public const uint kNumPosModels = kEndPosModelIndex - kStartPosModelIndex;

		public const uint kNumFullDistances = 1 << ((int)kEndPosModelIndex / 2);

		public const uint kNumLitPosStatesBitsEncodingMax = 4;
		public const uint kNumLitContextBitsMax = 8;

		public const int kNumPosStatesBitsMax = 4;
		public const uint kNumPosStatesMax = (1 << kNumPosStatesBitsMax);
		public const int kNumPosStatesBitsEncodingMax = 4;
		public const uint kNumPosStatesEncodingMax = (1 << kNumPosStatesBitsEncodingMax);

		public const int kNumLowLenBits = 3;
		public const int kNumMidLenBits = 3;
		public const int kNumHighLenBits = 8;
		public const uint kNumLowLenSymbols = 1 << kNumLowLenBits;
		public const uint kNumMidLenSymbols = 1 << kNumMidLenBits;
		public const uint kNumLenSymbols = kNumLowLenSymbols + kNumMidLenSymbols +
				(1 << kNumHighLenBits);
		public const uint kMatchMaxLen = kMatchMinLen + kNumLenSymbols - 1;
	}
}
